import { Language, LexCallback, LexRule, LexStateLabel } from "../def/language";
export declare type LexRuleLabel = string;
/**
 * 解析中の字句解析器の状態を操作するクラス
 */
export declare class LexController {
    private _lex;
    private _current_state;
    private _state_stack;
    private _rules;
    constructor(language: Language);
    /**
     * 個別にコールバックが設定されていない規則に対して適用するデフォルトコールバックを得る
     * @returns {LexCallback | undefined} デフォルトコールバック 定義されていない場合はundefined
     */
    readonly defaultCallback: LexCallback | undefined;
    /**
     * 字句解析開始時のコールバックを呼び出す
     */
    onBegin(): void;
    /**
     * 字句解析終了時のコールバックを呼び出す
     */
    onEnd(): void;
    /**
     * 現在の状態で適用可能な字句ルールをイテレータとして返す
     * @returns {IterableIterator<LexRule>} 字句ルールを得ることができるイテレータ
     */
    getRulesItr(): IterableIterator<LexRule>;
    /**
     * 新しい字句ルールを名前をつけて追加する
     * @param {string} label ルールの区別のために与える名前
     * @param {LexRule} rule 追加する字句ルール
     */
    addRule(label: string, rule: LexRule): void;
    /**
     * 既存の字句ルールを削除する
     * @param {string} label 削除するルールの名前
     * @returns {LexRule | undefined} 削除したルール 該当するものがない場合はundefined
     */
    removeRule(label: string): LexRule | undefined;
    /**
     * 現在の字句解析器の状態名を得る
     * @returns {LexStateLabel} 現在の状態名
     */
    getCurrentState(): LexStateLabel;
    /**
     * 字句解析機の解析状態を別の状態に変更する
     * @param {LexStateLabel} label 新しい状態の名前
     */
    jumpState(label: LexStateLabel): void;
    /**
     * 現在の状態をスタックに積んでから別の状態に変更する
     * @param {LexStateLabel} label 新しい状態の名前
     */
    callState(label: LexStateLabel): void;
    /**
     * スタックから1つ取り出し、その状態に変更する
     * スタックが空の場合は状態を変更しない
     * @returns {LexStateLabel | undefined} 変更した状態の名前 スタックが空の場合はundefined
     */
    returnState(): LexStateLabel | undefined;
}
