import { ILexer } from "../lexer/lexer";
import { Language } from "../def/language";
import { ASTNode } from "./ast";
/**
 * 字句解析・構文解析時に呼び出されるコールバックを管理するコントローラー
 * いわゆるStrategyパターン
 */
export interface CallbackController {
    readonly language: Language;
    init(): void;
    callLex(id: number, value: any, lexer: ILexer): any;
    callGrammar(id: number, children: Array<any>, lexer: ILexer): any;
}
/**
 * コールバックコントローラーの抽象クラス
 * 以下のメソッドをオーバーライドして使用する
 * * [[init]] (定義しなくてもよい)
 * * [[callLex]]
 * * [[callGrammar]]
 */
export declare abstract class AbstractCallbackController implements CallbackController {
    private _language;
    readonly language: Language;
    constructor(_language: Language);
    /**
     * 解析を開始する際、初期化のために呼び出される
     * デフォルトでは何もしない
     */
    init(): void;
    /**
     * 字句規則にマッチした際に呼び出される
     * @param {number} id
     * @param value
     * @param {ILexer} lexer
     * @returns {any}
     */
    abstract callLex(id: number, value: any, lexer: ILexer): any;
    abstract callGrammar(id: number, children: Array<any>, lexer: ILexer): any;
}
/**
 * 言語情報に付随するコールバックを呼び出すコントローラ
 */
export declare class DefaultCallbackController extends AbstractCallbackController {
    callLex(id: number, value: any, lexer: ILexer): any;
    callGrammar(id: number, children: Array<any>, lexer: ILexer): any;
}
/**
 * ASTを構築するコントローラ
 */
export declare class ASTConstructor extends AbstractCallbackController {
    callLex(id: number, value: any, lexer: ILexer): ASTNode;
    callGrammar(id: number, children: Array<any>, lexer: ILexer): ASTNode;
}
