"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const token_1 = require("../def/token");
const nullableset_1 = require("./nullableset");
/**
 * First集合
 */
class FirstSet {
    /**
     * @param {GrammarDefinition} grammar 構文規則
     * @param {SymbolDiscriminator} symbols 終端/非終端記号の判別に用いる分類器
     */
    constructor(grammar, symbols) {
        this.grammar = grammar;
        this.symbols = symbols;
        this.first_map = new Map();
        this.nulls = new nullableset_1.NullableSet(this.grammar);
        this.generateFirst();
    }
    /**
     * First集合を生成する
     */
    generateFirst() {
        // Firstを導出
        const first_result = new Map();
        // 初期化
        // FIRST($) = {$} だけ手動で追加
        first_result.set(token_1.SYMBOL_EOF, new Set([token_1.SYMBOL_EOF]));
        // 終端記号Xに対してFirst(X)=X
        const terminal_symbols = this.symbols.getTerminalSymbols();
        terminal_symbols.forEach((value) => {
            first_result.set(value, new Set([value]));
        });
        // 非終端記号はFirst(Y)=∅で初期化
        const nonterminal_symbols = this.symbols.getNonterminalSymbols();
        nonterminal_symbols.forEach((value) => {
            first_result.set(value, new Set());
        });
        // 包含についての制約を生成
        const constraint = [];
        for (const rule of this.grammar.rules) {
            const sup = rule.ltoken;
            // 右辺の左から順に、non-nullableな記号が現れるまで制約に追加
            // 最初のnon-nullableな記号は制約に含める
            for (const sub of rule.pattern) {
                if (sup != sub) {
                    constraint.push({ superset: sup, subset: sub });
                }
                if (!this.nulls.isNullable(sub)) {
                    break;
                }
            }
        }
        // 制約解消
        let flg_changed = true;
        while (flg_changed) {
            flg_changed = false;
            for (const pair of constraint) {
                const sup = pair.superset;
                const sub = pair.subset;
                const superset = first_result.get(sup);
                const subset = first_result.get(sub);
                subset.forEach((token) => {
                    // subset内の要素がsupersetに含まれていない
                    if (!superset.has(token)) {
                        // subset内の要素をsupersetに入れる
                        superset.add(token);
                        flg_changed = true;
                    }
                });
                // First集合を更新
                first_result.set(sup, superset);
            }
        }
        this.first_map = first_result;
    }
    /**
     * 記号または記号列を与えて、その記号から最初に導かれうる非終端記号の集合を返す
     * @param {Token | Token[]} arg
     * @returns {Set<Token>}
     */
    get(arg) {
        // 単一の記号の場合
        if (!Array.isArray(arg)) {
            if (!this.first_map.has(arg)) {
                throw new Error(`invalid token found: ${arg}`);
            }
            return this.first_map.get(arg);
        }
        // 記号列の場合
        const tokens = arg;
        // 不正な記号を発見
        for (const token of tokens) {
            if (!this.first_map.has(token)) {
                throw new Error(`invalid token found: ${token}`);
            }
        }
        const result = new Set();
        for (const token of tokens) {
            const add = this.first_map.get(token); // トークン列の先頭から順にFirst集合を取得
            // 追加
            add.forEach((t) => {
                if (!result.has(t)) {
                    result.add(t);
                }
            });
            if (!this.nulls.isNullable(token)) {
                // 現在のトークン ∉ Nulls ならばここでストップ
                break;
            }
        }
        return result;
    }
}
exports.FirstSet = FirstSet;
