"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * 終端/非終端記号の判別を行う
 */
class SymbolDiscriminator {
    constructor(grammar) {
        this.terminal_symbols = new Set();
        this.nonterminal_symbols = new Set();
        // 左辺値の登録
        for (const rule of grammar.rules) {
            const symbol = rule.ltoken;
            // 構文規則の左辺に現れる記号は非終端記号
            this.nonterminal_symbols.add(symbol);
        }
        // 右辺値の登録
        for (const rule of grammar.rules) {
            for (const symbol of rule.pattern) {
                if (!this.nonterminal_symbols.has(symbol)) {
                    // 非終端記号でない(=左辺値に現れない)場合、終端記号である
                    this.terminal_symbols.add(symbol);
                }
            }
        }
    }
    /**
     * 終端記号の集合をSetで得る
     * @param {boolean} prevent_copy trueを与えるとSetをコピーせず返す
     * 結果に変更が加えられないと保証される場合に用いる
     * @returns {Set<Token>}
     */
    getTerminalSymbols(prevent_copy = false) {
        if (prevent_copy)
            return this.terminal_symbols;
        // コピーを返す
        return new Set(this.terminal_symbols);
    }
    /**
     * 非終端記号の集合をSetで得る
     * @param {boolean} prevent_copy trueを与えるとSetをコピーせず返す
     * 結果に変更が加えられないと保証される場合に用いる
     * @returns {Set<Token>}
     */
    getNonterminalSymbols(prevent_copy = false) {
        if (prevent_copy)
            return this.nonterminal_symbols;
        // コピーを返す
        return new Set(this.nonterminal_symbols);
    }
    /**
     * 与えられた記号が終端記号かどうかを調べる
     * @param {Token} symbol
     * @returns {boolean}
     */
    isTerminalSymbol(symbol) {
        return this.terminal_symbols.has(symbol);
    }
    /**
     * 与えられた記号が非終端記号かどうかを調べる
     * @param {Token} symbol
     * @returns {boolean}
     */
    isNonterminalSymbol(symbol) {
        return this.nonterminal_symbols.has(symbol);
    }
}
exports.SymbolDiscriminator = SymbolDiscriminator;
